#!/usr/bin/env python

# Copyright (c) 2011 Red Hat, Inc.
#
# This file is part of HekaFS.
#
# HekaFS is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License, version 3, as published by the Free
# Software Foundation.
#
# HekaFS is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License * along
# with HekaFS.  If not, see <http://www.gnu.org/licenses/>.

import json
import optparse
import os
import sys
import urllib2

import hfs_paths
import hfs_utils
import volfilter

# Simple cache of brick-to-port mappings, so we don't have to keep re-fetching
# the maps from the same host if it has multiple bricks.
class mapper:
	def __init__ (self):
		self.cache = {}
	def lookup (self, host, volume, subv):
		if self.cache.has_key(host):
			mydict = self.cache[host]
		else:
			url = "http://%s:%d/%s/map" % \
				(host, hfs_paths.HEKAFSD_PORT, volume)
			mydict = json.load(urllib2.urlopen(url))
			self.cache[host] = mydict
		if mydict.has_key(subv):
			return mydict[subv]


if __name__ == "__main__":

	cache = mapper()

	my_usage = "%prog [options] server volume user pw mountpoint"
	parser = optparse.OptionParser(usage=my_usage)
	parser.add_option("-d","--data-key",dest="keyfile",default="",
		help="encryption key-file for on-disk encryption")
	parser.add_option("-p","--private-key",dest="private_key",default="",
		help="private key for SSL")
	parser.add_option("-c","--client-cert",dest="client_cert",default="",
		help="client's own certificate for SSL")
	parser.add_option("-s","--server-cert",dest="server_cert",default="",
		help="server-certificate list for SSL")
	options, args = parser.parse_args()

	ssl_opts = 0
	for oname in ("private_key", "client_cert", "server_cert"):
		if getattr(options,oname):
			ssl_opts += 1
	if (ssl_opts > 0) and (ssl_opts < 3):
		print "-p/-c/-s must be specified together"
		sys.exit(1)

	if len(args) != 5:
		parser.print_help()
		sys.exit(1)
	(host, volume, username, password, mount) = args

	# Fetch the GlusterFS client-side volfile.
	url = "http://%s:%d/%s/fetch" % (host, hfs_paths.HEKAFSD_PORT, volume)
	vol_file = urllib2.urlopen(url)

	# Load the volfile and clean out some of the crud.
	graph, last = volfilter.load(vol_file)
	last = volfilter.cleanup(last,graph)

	# For each client translator, map to the port that hekafsd allocated
	# for that brick and add our username/password to select the tenant.
	for xl in graph.itervalues():
		if xl.type != "protocol/client2":
			continue
		host = xl.opts["remote-host"]
		subv = xl.opts["remote-subvolume"]
		port = cache.lookup(host,volume,subv)
		if not port:
			errstr = "Could not find port for %s:%s" % (host, subv)
			print >> sys.stderr, errstr
			sys.exit(1)
		xl.opts["remote-port"] = port
		xl.opts["remote-subvolume"] = username + "-uidmap"
		xl.opts["username"] = username
		xl.opts["password"] = password
		if ssl_opts:
                        xl.opts["transport.socket.ssl-enabled"] = "true"
			xl.opts["transport.socket.ssl-private-key"] \
				= options.private_key
			xl.opts["transport.socket.ssl-own-cert"] \
				= options.client_cert
			xl.opts["transport.socket.ssl-ca-list"] \
				= options.server_cert

	if options.keyfile != "":
                to_do = [x for x in graph.itervalues()
                                if x.type == "cluster/replicate"]
		for xl in to_do:
                        volfilter.push_filter(graph,xl,
                                "encryption/crypt",
                                {"key":options.keyfile})

	outfile = os.path.join(hfs_paths.info_dir,"%s.vol"%volume)
	volfilter.generate(graph,last,open(outfile,"w"))
	cmd = "-f %s %s" % (outfile, mount)
	print "running glusterfs %s" % cmd
	hfs_utils.run_cmd("glusterfs",cmd).wait()




