#!/usr/bin/env python

# Copyright (c) 2011 Red Hat, Inc.
#
# This file is part of HekaFS.
#
# HekaFS is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License, version 3, as published by the Free
# Software Foundation.
#
# HekaFS is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License * along
# with HekaFS.  If not, see <http://www.gnu.org/licenses/>.

import optparse
import glob
import os
import re
import socket
import string
import subprocess
import sys
import urllib2

from bottle import template

import volfilter
import hfs_paths
import hfs_utils

# Make sure the volume directory exists and has the right stuff in it.
def check_volume_directory(vol_name):
	if not os.path.exists(hfs_paths.info_dir):
		os.mkdir(hfs_paths.info_dir,0700)
	if not os.path.exists(hfs_paths.log_dir):
		os.mkdir(hfs_paths.log_dir,0700)
	if not os.path.exists(hfs_paths.pid_dir):
		os.mkdir(hfs_paths.pid_dir,0700)
	if not os.path.exists(hfs_paths.idle_subdir):
		os.mkdir(hfs_paths.idle_subdir,0700)
	if not os.path.exists(hfs_paths.used_subdir):
		os.mkdir(hfs_paths.used_subdir,0700)
	idle = glob.glob("%s/*" % hfs_paths.idle_subdir)
	used = glob.glob("%s/*" % hfs_paths.used_subdir)
	if len(idle) + len(used) == 0:
		# idle and used dirs are both empty, probably after a reboot
		# when /run is cleaned of files and symlinks
		# creating idle ports
		for i in range(24010, 24030):
			print "creating %s/%d" % (hfs_paths.idle_subdir, i)
			fp = open("%s/%d" % (hfs_paths.idle_subdir, i), "w")
			fp.close()

	vol_dir = "%s/%s" % (hfs_paths.info_dir, vol_name)
	if not os.path.exists(vol_dir):
		os.mkdir(vol_dir)
	return vol_dir

# Figure out which of the GlusterFS volfiles belong to us.  Volfiles might have
# names based on partial host names, fully qualified names, or addresses, or
# even a mix thanks to "gluster peer probe" silliness.  To deal with all of
# these possibilities, we resolve everything to addresses and compare those.
# ### bear in mind that depending on how a machine is set up, the IP addrs
# ### for a node might include 127.0.0.1 and ::1 first
def scan_gfs_volfiles(vol_name):
        ret = []
	my_name = os.uname()[1]
	# Getaddrinfo returns a list of tuples, each:
	#	family, socktype, proto, canonname, sockaddr
	# We extract the sockaddr of the first item, and the IP addr from that
	# TBD: handle IPv6, multi-homed hosts, etc.
	# TBD: skip loopback addresses based on note above
	my_addrs = socket.getaddrinfo(my_name, 0)
	my_glob = "%s/vols/%s/%s.*.vol" % (hfs_paths.gfs_dir, vol_name, vol_name)
	for vf in glob.iglob(my_glob):
		m = hfs_paths.volfile_re.match(os.path.basename(vf))
		if m:
			this_host = m.group("node")
			this_addr = socket.getaddrinfo(this_host, 0)[0][4][0]
			for addr in my_addrs:
				if this_addr == addr[4][0]:
					ret.append(vf)
					break
	return ret

# Allocate a port for a server to run on.  Right now we do this in a very
# "clever" way, by creating files to match ports and then grabbing a file here.
# When we have a real volume database such games will be unnecessary.
def allocate_port(vol_file):
	for pf in glob.iglob("%s/*" % hfs_paths.idle_subdir):
		base = os.path.basename(pf)
		new_name = "%s/%s" % (hfs_paths.used_subdir, base)
		os.remove(pf)
		os.symlink(vol_file, new_name)
		return base
	else:
		raise RuntimeError, "no ports available"

# Parse the user file into a list of [name,password] sub-lists.  Since
# everything that uses this is in Python we could just make it a pickle/shelf
# or whatever, but it would all go away with a real volume database so it's not
# worth the trouble to re-do it now.
def parse_user_file (vol_name):
	db_obj = hfs_utils.open_db()
	users = []
	user_list = db_obj["vt_"+vol_name].split(",")
	if user_list == ['']:
		user_list = []
	for name in user_list:
		print "found user %s" % name
		pw = db_obj["tp_"+name]
		ul = db_obj["tul_"+name]
		uh = db_obj["tuh_"+name]
		gl = db_obj["tgl_"+name]
		gh = db_obj["tgh_"+name]
		users.append([name,pw,ul,uh,gl,gh])
	return users

def cloudify_server (input, output, users, port, ca_path):
	print "# Cloudifying server %s" % input
	graph, last = volfilter.load(input)
	last = volfilter.cleanup(last,graph)

	if last.type != "protocol/server2":
		print >> sys.stderr, "Top translator must be protocol/server2"
		sys.exit(1)
	if ca_path:
                last.opts["transport.socket.ssl-enabled"] = "true";
		last.opts["transport.socket.ssl-private-key"] = \
			os.path.join(hfs_paths.info_dir,"server.key")
		last.opts["transport.socket.ssl-own-cert"] = \
			os.path.join(hfs_paths.info_dir,"server.pem")
		last.opts["transport.socket.ssl-ca-list"] = ca_path
		ca_file = open(ca_path,"w+")
	old_stack = last.subvols[0]

	bad_opts = []
	for opt in last.opts.iterkeys():
		if opt[:9] == "auth.addr":
			bad_opts.append(opt)
		elif opt[:10] == "auth.login":
			bad_opts.append(opt)
	for opt in bad_opts:
		print "# stripping auth option %s = %s" % (opt, last.opts[opt])
		del last.opts[opt]

	uidmap_opts = dict()
	last.subvols = []
	for user, pw, ul, uh, gl, gh in users:
		if ca_path:
			cert_path = os.path.join(hfs_paths.info_dir,
						 "client_%s.pem" % user)
			if not os.path.isfile(cert_path):
				print "skipping %s (no cert)" % user
				continue
			new_cert = open(cert_path,"r").read()
			ca_file.write(new_cert)
		uidmap_opts[user] = dict()
		uidmap_opts[user]["uidmap-plugin"] = "libmaprbtree.so"
		uidmap_opts[user]["uid-range"] = "%s-%s" % (ul, uh)
		uidmap_opts[user]["gid-range"] = "%s-%s" % (gl, gh)
		new_stack = volfilter.copy_stack(old_stack,user)
		volfilter.push_filter(graph, new_stack, "features/uidmap", uidmap_opts[user])
		last.subvols.append(new_stack)
		last.opts["auth.login.%s.allow"%new_stack.name] = user
		last.opts["auth.login.%s.password"%user] = pw 
	if ca_path:
		ca_file.flush()

	last.opts["transport.socket.listen-port"] = port
	volfilter.generate(graph,last,output)

def create_tenant_dirs(vol_file):
	cmd = "/bin/grep \"option directory\" %s" % vol_file
	path = ""
	opt_dir_lines = os.popen(cmd)
	for opt_dir_line in opt_dir_lines:
		tokens = re.split(' ', string.lstrip(opt_dir_line))
		path = string.rstrip(tokens[2])
		ppath = os.path.dirname(path)
		if not os.path.exists(ppath):
			os.mkdir(ppath)
		if not os.path.exists(path):
			os.mkdir(path)
			# temporary, see BZ 747112; this is the /tmp model
			os.chmod(path, 01777)
	opt_dir_lines.close()

def start_local (vol_name):
	vol_base = check_volume_directory(vol_name)
	users = parse_user_file(vol_name)
	if not len(users):
		return ["no users for %s on %s" % (
			vol_name, socket.gethostname())]
	results = []
	for vf in scan_gfs_volfiles(vol_name):
		new_vf = os.path.join(vol_base, os.path.basename(vf))
		outfile = open(new_vf, "w")
		port = allocate_port(new_vf)
		print "got port %s for %s" % (port, new_vf)
		db_obj = hfs_utils.open_db()
		if db_obj.get("vs_"+vol_name):
			ca_path = os.path.join(hfs_paths.info_dir,
					       "vol_%s.ca" % vol_name)
		else:
			ca_path = None
		cloudify_server(vf, outfile, users, port, ca_path)
		if ca_path:
			print "generated CA list %s" % ca_path
		outfile.flush()
		outfile.close()
		# print "v_key: %s" % v_key
		# make dirs for each of the users
		create_tenant_dirs(new_vf)
		# actually start the server so that gluster can find it
		# note: pid file in /var/lib/glusterd/vols/<vol_name>/...
		v_key = string.replace(os.path.basename(new_vf), ".vol", "")
		# print "v_key: %s" % v_key
		logfile = os.path.join(hfs_paths.log_dir,"%s.log"%v_key)
		pidfile = os.path.join(hfs_paths.pid_dir,"%s.pid"%v_key)
		cmd = "--volfile %s" % new_vf
		cmd += (" --log-file %s" % logfile)
		cmd += (" --pid-file %s" % pidfile)
		cmd += " --xlator-option"
		cmd += (" %s-server.transport.socket.listen-port=%s" % (
			vol_name, port))
		#print "whole command = glusterfsd %s" % cmd
		retcode = hfs_utils.run_cmd("glusterfsd",cmd).wait()
		vol_result = "start_local(%s) returned %d on %s\n" % (
			v_key, retcode, socket.gethostname())
		results.append(vol_result)
	return results

def run_common (vol_name):
	node_list = hfs_utils.get_nodes_for_vol(vol_name)
	blob = []
	for node in node_list:
		scratch = [node, []]
		if socket.gethostbyname(node) in hfs_utils.local_addrs:
			url_obj = start_local(vol_name)
		else:
			url = "http://%s:%d/volumes/%s/start_local" % (
				node, hfs_paths.HEKAFSD_PORT, vol_name)
			url_obj = urllib2.urlopen(url)
		for line in url_obj:
			scratch[1].append(line)
		blob.append(scratch)
	return blob

def run_www (vol_name):
	blob = run_common(vol_name)
	return template("start_done.html",name=vol_name,blob=blob)

if __name__ == "__main__":
	my_usage = "%prog volume"
	parser = optparse.OptionParser(usage=my_usage)
	options, args = parser.parse_args()
	if len(args) != 1:
		parser.print_help()
		sys.exit(1)
	name = args[0]
	blob = run_common(name)
	print "Volume %s started." % name
	hfs_utils.print_blob(blob)

