# Copyright (c) 2011 Red Hat, Inc.
#
# This file is part of HekaFS.
#
# HekaFS is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License, version 3, as published by the Free
# Software Foundation.
#
# HekaFS is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License * along
# with HekaFS.  If not, see <http://www.gnu.org/licenses/>.

import dbm
import glob
import os
import socket
import subprocess

import hfs_paths

# The list of filesystems that work is much shorter than the list of
# filesystems that don't support xattrs, already-remote filesystems (e.g. NFS),
# pseudo-filesystems (e.g. devfs) or other things that won't work for one
# reason or another.
good_fs_types = [ "ext2", "ext3", "ext4", "xfs", "btrfs" ]

# some mount points reasonably certain not used for bricks
disqualified_mounts = [ "/boot", "/home", "/tmp", "/var/tmp", "/" ]

# Sudo is kind of pointless if we're already running as root.
use_sudo = False

# We use the class cache to avoid having to call through the shell to find
# executables on $PATH every single time.  To do this, we use the shell *once*
# to find the local path and assume all nodes are configured similarly.
class_cache = {}

def get_path (program):
	if not class_cache.has_key(program):
		for dir in os.getenv("PATH").split(":"):
			maybe = os.path.join(dir,program)
			if os.access(maybe,os.X_OK):
				class_cache[program] = maybe
				break
	# KeyError is as good as any exception we might throw ourselves.
	return class_cache[program]

# This lets users change access methods (e.g. pdsh, slurm, some future method
# that proxies the command through hekafsd's own secure connection.
def make_remote (host, cmd):
	return "%s %s %s" % (get_path("ssh"), host, cmd)

def run_cmd (program, args, host=None, filters=[]):
	cmd = "%s %s" % (get_path(program), args)
	for prog2, args2 in filters:
		cmd = "%s | %s %s" % (cmd, get_path(prog2), args2)
	if use_sudo:
		cmd = "sudo %s" % cmd
	if host and host not in local_addrs:
		cmd = make_remote(host,cmd)
	if len(filters):
		#print "executing %s using shell" % cmd
		child = subprocess.Popen(cmd,close_fds=True,shell=True,
					 stdout=subprocess.PIPE)
	else:
		#print "executing %s without shell" % cmd
		cmd = cmd.split()
		child = subprocess.Popen(cmd,close_fds=True,shell=False,
					 stdin=subprocess.PIPE,
					 stdout=subprocess.PIPE)
	return child

# Figure out what our local IP addresses are, for future reference.
# TBD: IPv6 yadda yadda
local_addrs = []
for line in run_cmd("ip","addr").stdout:
	parts = line.lstrip().split(" ")
	if parts[0] == "inet":
		addr = parts[1]
		if not addr.startswith("127."):
			local_addrs.append(addr.split("/")[0])
#print local_addrs

# NB this node is always first in the list
def get_members ():
	peer_ips = run_cmd("gluster","peer status",
			   filters=[("grep","Hostname:")]).stdout
	node_list = [socket.gethostbyname(socket.gethostname())]
	# For each line: strip off the trailing newline, split at colon, take
	# item #1
	for peer in [line[:-1].split(": ")[1] for line in peer_ips]:
		node_list.append(peer)
	node_list.sort()
	return node_list

# A brick list looks something like this:
# 	{
#		"vol_A": [ "server1:/bricks/A", "server2:/bricks/A" ],
#		"vol_B": [ "server2:/bricks/B", "server3:/bricks/B: ]
#	}
def get_bricks ():
	vol_list = {}
	brick_list = []
	for line in run_cmd("gluster","volume info all").stdout:
		parts = line[:-1].split(": ")
		if len(parts) != 2:
			continue
		if parts[0] == "Volume Name":
			if brick_list:
				vol_list[vol_name] = brick_list
			vol_name = parts[1]
			brick_list = []
		elif parts[0].startswith("Brick"):
			brick_list.append(parts[1])
	if brick_list:
		vol_list[vol_name] = brick_list
	return vol_list

def dir_in_use (brick_list, dir):
	for vol_bricks in brick_list.itervalues():
		for brick in vol_bricks:
			if brick == dir:
				return True
	return False

# A mount list looks something like this:
#	{
#		"server1": [ ( "/bricks/A", True ),
#			     ( "/some/other/mount", False) ]
#		"server2": [ ( "/bricks/A", True ),
#			     ( "/bricks/B", True ) ],
#		"server3": [ ( "/bricks/B", True ),
#			     ( "/what/ever", False ) ]
#	}
# The second part of each tuple is the "in-use" flag, true iff the mount is
# part of a volume.
def get_mounts (brick_list):
	volumes_on_nodes = {}
	for node_ip in get_members() :
		scratch = []
		for line in run_cmd("mount","",host=node_ip).stdout:
			bits = line.split(" ")
			if bits[4] not in good_fs_types or bits[2] in disqualified_mounts:
				continue
			mount = "%s:%s" % (node_ip, bits[2])
			inuse = dir_in_use(brick_list,mount)
			scratch.append((bits[2],inuse))
		scratch.sort()
		volumes_on_nodes[node_ip] = scratch
	# Add user-defined non-mountpoint directories.
	db_obj = open_db()
	for node in [k[3:] for k in db_obj.keys() if k.startswith("xp_")]:
		dirs = db_obj["xp_"+node].split(",")
		if dirs == [""]:
			dirs = []
		for dir in dirs:
			mount = "%s:%s" % (node, dir)
			inuse = dir_in_use(brick_list,mount)
			volumes_on_nodes[node].append((dir,inuse))
	return volumes_on_nodes

# Get the list of nodes that are serving (any part of) a specific volume.
def get_nodes_for_vol (vol_name):
	node_list = set()
	my_glob = "%s/vols/%s/%s.*.vol" % (hfs_paths.gfs_dir,
		vol_name, vol_name)
	for vf in glob.iglob(my_glob):
		m = hfs_paths.volfile_re.match(os.path.basename(vf))
		if m:
			node_list.add(m.group("node"))
	return node_list

# Open our configuration database.
# We use dbm explicitly because we don't want to get into a situation where
# a DB was created using some more "advanced" library and then restored/moved
# without that library present.
def open_db ():
	db_path = os.path.join(hfs_paths.info_dir,"config.db")
	return dbm.open(db_path,"c",0600)

# Print a "blob" of [node, [line1, line2]] tuples/lists for CLI/debugging.
def print_blob (blob):
	for node, text in blob:
		print "= %s =" % node
		for line in text:
			print line
